# -*- coding: latin-1 -*-
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os.path
from unittest import mock

import tempfile
import unittest

from monasca_agent.common import config
from monasca_agent.common.config import Config
from monasca_agent.common import util


class TestConfig(unittest.TestCase):

    def xtestWhiteSpaceConfig(self):
        """Leading whitespace confuse ConfigParser
        """
        agent_config = Config.get_config(
            cfg_path=os.path.join(os.path.dirname(os.path.realpath(__file__)), "badconfig.conf"))
        self.assertEqual(agent_config["api_key"], "1234")

    def testGoodPidFie(self):
        """Verify that the pid file succeeds and fails appropriately"""

        pid_dir = tempfile.mkdtemp()
        program = 'test'

        expected_path = os.path.join(pid_dir, '%s.pid' % program)
        pid = "666"
        pid_f = open(expected_path, 'w')
        pid_f.write(pid)
        pid_f.close()

        p = util.PidFile(program, pid_dir)

        self.assertEqual(p.get_pid(), 666)
        # clean up
        self.assertEqual(p.clean(), True)
        self.assertEqual(os.path.exists(expected_path), False)

    def testHostname(self):
        valid_hostnames = [
            u'i-123445',
            u'5dfsdfsdrrfsv',
            u'432498234234A'
            u'234234235235235235',  # Couldn't find anything in the RFC saying it's not valid
            u'A45fsdff045-dsflk4dfsdc.ret43tjssfd',
            u'4354sfsdkfj4TEfdlv56gdgdfRET.dsf-dg',
            u'r' * 255,
        ]

        not_valid_hostnames = [
            u'abc' * 150,
            u'sdf4..sfsd',
            u'$42sdf',
            u'.sfdsfds'
            u's™£™£¢ª•ªdfésdfs'
        ]

        for hostname in valid_hostnames:
            self.assertTrue(util.is_valid_hostname(hostname), hostname)

        for hostname in not_valid_hostnames:
            self.assertFalse(util.is_valid_hostname(hostname), hostname)

    def testConfigIsSingleton(self):
        # create a temp conf file
        tempdir = tempfile.mkdtemp()
        conf_file = os.path.join(tempdir, 'agent.yaml')
        with open(conf_file, 'wb') as fd:
            fd.write(
                b"""
                Logging:
                  collector_log_file: /var/log/monasca/agent/collector.log
                  forwarder_log_file: /var/log/monasca/agent/forwarder.log
                  log_level: DEBUG
                  statsd_log_file: /var/log/monasca/agent/statsd.log
                Main:
                  check_freq: 60
                  dimensions: {}
                  hostname: example.com
                """
            )
        conf_1 = Config(configFile=conf_file)
        conf_2 = Config(configFile=conf_file)
        conf_3 = Config()
        self.assertTrue(conf_1 is conf_2)
        self.assertTrue(conf_1 is conf_3)

    @mock.patch.object(Config, '_read_config')
    @mock.patch('monasca_agent.common.util.get_parsed_args')
    def testConfigFromParsedArgs(self, mock_parsed_args, mock_read_config):
        mock_options = mock.Mock()
        mock_parsed_args.return_value = (mock_options, mock.sentinel.args)
        conf = Config()
        # object is singleton, for this case, it needs to be reloaded.
        conf.__init__()

        self.assertEqual(mock_options.config_file, conf._configFile)

    @mock.patch.object(Config, '_read_config')
    @mock.patch('monasca_agent.common.util.get_parsed_args')
    @mock.patch('monasca_agent.common.config.os')
    def testConfigFileFromDefault(self, mock_os, mock_parsed_args, mock_read_config):
        mock_os.path.exists = mock.create_autospec(mock_os.path.exists, return_value=True)
        mock_options = mock.Mock()
        mock_options.config_file = None
        mock_parsed_args.return_value = (mock_options, mock.sentinel.args)
        conf = Config()
        # object is singleton, for this case, it needs to be reloaded.
        conf.__init__()

        self.assertEqual(config.DEFAULT_CONFIG_FILE, conf._configFile)
        mock_os.path.exists.assert_called_once_with(config.DEFAULT_CONFIG_FILE)

    def test_verify_common_config_opts(self):
        opts = util.get_parsed_args(prog='test')
        opts_dict = vars(opts[0])
        self.assertCountEqual(['config_file', 'clean', 'verbose'],
                             opts_dict.keys())



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import platform
import sys

from monasca_setup.service import linux


LOG = logging.getLogger(__name__)


def detect_init(*args, **kwargs):
    """Create a service object if possible.

    Detect if systemd is present on the system and if so return the service
    object.

    :return: a systemd service object for this system.
    """
    detected_os = platform.system()
    if has_systemd():
        return linux.Systemd(*args, **kwargs)
    LOG.error("{0} is not currently supported by the Monasca Agent"
              .format(detected_os))
    sys.exit(1)


def has_systemd():
    """Detect if Linux init is systemd."""
    with open('/proc/1/comm', 'r') as init_proc:
        init = init_proc.readline().strip()
        return init == 'systemd'



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

""" Systemd based service
"""
import logging
import os
import pwd
import subprocess

LOG = logging.getLogger(__name__)


class Systemd(object):
    """Manage service using systemd."""

    def __init__(self, prefix_dir, config_dir, log_dir, template_dir,
                 name='monasca-agent', username='mon-agent'):
        """Create a service."""
        self.prefix_dir = prefix_dir
        self.config_dir = config_dir
        self.log_dir = log_dir
        self.template_dir = template_dir
        self.name = name
        self.username = username

    def enable(self):
        """Set monasca-agent to start on boot.

        Generally this requires running as super user.
        """
        if os.geteuid() != 0:
            LOG.error('This service must be run as root')
            raise OSError

        # LinuxInit.enable(self)
        # Create user/group if needed
        try:
            user = pwd.getpwnam(self.username)
        except KeyError:
            subprocess.check_call(['useradd', '-r', self.username])
            user = pwd.getpwnam(self.username)

        # Create dirs
        for path in (self.log_dir, self.config_dir,
                     '%s/conf.d' % self.config_dir):
            if not os.path.exists(path):
                os.makedirs(path, 0o755)
                os.chown(path, 0, user.pw_gid)
        # log dir needs to be writable by the user
        os.chown(self.log_dir, user.pw_uid, user.pw_gid)

        # Get systemd services and target template
        templates = [f for f in os.listdir(self.template_dir)
                     if (f.endswith('service.template') or
                         f.endswith('target.template'))]
        systemd_path = '/etc/systemd/system/'

        # Write the systemd units configuration file: we have 3 services and
        # 1 target grouping all of them together
        for template_file_name in templates:
            service_file_name, e = os.path.splitext(template_file_name)
            service_file_path = os.path.join(systemd_path,
                                             service_file_name)
            with open(os.path.join(self.template_dir,
                      template_file_name), 'r') as template:
                with open(service_file_path, 'w') as service_file:
                    LOG.info('Creating service file %s', service_file_name)
                    service_file.write(template.read().
                                       format(prefix=self.prefix_dir,
                                              monasca_user=self.username))
            os.chown(service_file_path, 0, 0)
            os.chmod(service_file_path, 0o644)

        # Enable the service
        subprocess.check_call(['systemctl', 'daemon-reload'])
        subprocess.check_call(
            ['systemctl', 'enable', '{0}.target'.format(self.name)])
        LOG.info('Enabled %s target via systemd', self.name)

    def start(self, restart=True):
        """Start monasca-agent.

        If the agent is running and restart is True restart it.

        :return: True if monasca-agent is enabled on boot, False otherwise..
        """
        if not self.is_enabled():
            LOG.error('The service is not enabled')
            return False

        LOG.info('Starting %s services via systemd', self.name)
        if self.is_running() and restart:
            subprocess.check_call(
                ['systemctl', 'restart', '{0}.target'.format(self.name)])
        else:
            subprocess.check_call(
                ['systemctl', 'start', '{0}.target'.format(self.name)])
        return True

    def stop(self):
        """Stop monasca-agent.
        :return: True if monasca-agent was stopped successfully, False otherwise
        """
        LOG.info('Stopping %s services', self.name)
        try:
            subprocess.check_call(
                ['systemctl', 'stop', '{0}.target'.format(self.name)])
        except subprocess.CalledProcessError as call_error:
            LOG.error('Unable to stop monasca-agent.')
            LOG.error(call_error.output)
            return False
        else:
            return True

    def is_running(self):
        """Check if monasca-agent is running.

        :return: True if monasca-agent is running, false otherwise.
        """
        return(subprocess.call(['systemctl', 'is-active', '--quiet',
                               '{0}.target'.format(self.name)]) == 0)

    def is_enabled(self):
        """Check if monasca-agent is setup to start at boot time.

        :return: True if monasca-agent is enabled on boot, False otherwise.
        """
        return(subprocess.call(['systemctl', 'is-enabled', '--quiet',
                               '{0}.target'.format(self.name)]) == 0)



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import psutil

log = logging.getLogger(__name__)

import monasca_agent.collector.checks as checks


class Memory(checks.AgentCheck):

    def __init__(self, name, init_config, agent_config):
        super(Memory, self).__init__(name, init_config, agent_config)
        process_fs_path_config = init_config.get('process_fs_path', None)
        if process_fs_path_config:
            psutil.PROCFS_PATH = process_fs_path_config
            self.log.debug('The path of the process filesystem set to %s', process_fs_path_config)
        else:
            self.log.debug('The process_fs_path not set. Use default path: /proc')

    def check(self, instance):
        """Capture memory stats

        """
        dimensions = self._set_dimensions(None, instance)

        mem_info = psutil.virtual_memory()
        swap_info = psutil.swap_memory()

        self.gauge('mem.total_mb',
                   int(mem_info.total / 1048576),
                   dimensions=dimensions)
        self.gauge('mem.free_mb',
                   int(mem_info.free / 1048576),
                   dimensions=dimensions)
        self.gauge('mem.usable_mb',
                   int(mem_info.available / 1048576),
                   dimensions=dimensions)
        self.gauge('mem.usable_perc',
                   float(100 - mem_info.percent),
                   dimensions=dimensions)
        self.gauge('mem.swap_total_mb',
                   int(swap_info.total / 1048576),
                   dimensions=dimensions)
        self.gauge('mem.swap_used_mb',
                   int(swap_info.used / 1048576),
                   dimensions=dimensions)
        self.gauge('mem.swap_free_mb',
                   int(swap_info.free / 1048576),
                   dimensions=dimensions)
        self.gauge('mem.swap_free_perc',
                   float(100 - swap_info.percent),
                   dimensions=dimensions)

        count = 9
        if hasattr(mem_info, 'buffers') and mem_info.buffers:
            self.gauge('mem.used_buffers',
                       int(mem_info.buffers / 1048576),
                       dimensions=dimensions)
            count += 1

        if hasattr(mem_info, 'cached') and mem_info.cached:
            self.gauge('mem.used_cache',
                       int(mem_info.cached / 1048576),
                       dimensions=dimensions)
            count += 1

        if (hasattr(mem_info, 'buffers') and mem_info.buffers and
                hasattr(mem_info, 'cached') and mem_info.cached):

            mem_used_real = mem_info.used
            if psutil.version_info < (4, 4, 0):
                #
                # pusutil versions prior to 4.4.0 didn't subtract buffers and
                # cache, but starting in 4.4.0 psutil does.
                #
                mem_used_real = mem_used_real - mem_info.buffers - mem_info.cached
            self.gauge('mem.used_real_mb', int(mem_used_real / 1048576),
                       dimensions=dimensions)
            count += 1

        if hasattr(mem_info, 'shared') and mem_info.shared:
            self.gauge('mem.used_shared',
                       int(mem_info.shared / 1048576),
                       dimensions=dimensions)
            count += 1

        # The slab metric was added in psutil 5.4.4
        if hasattr(mem_info, 'slab') and mem_info.slab:
            self.gauge('mem.used_slab_mb',
                       int(mem_info.slab / 1048576),
                       dimensions=dimensions)
            count += 1

        log.debug('Collected {0} memory metrics'.format(count))



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import psutil
import re

log = logging.getLogger(__name__)

import monasca_agent.collector.checks as checks


class Disk(checks.AgentCheck):

    def __init__(self, name, init_config, agent_config):
        self._partition_error = set()

        super(Disk, self).__init__(name, init_config, agent_config)
        process_fs_path_config = init_config.get('process_fs_path', None)
        if process_fs_path_config:
            psutil.PROCFS_PATH = process_fs_path_config
            self.log.debug('The path of the process filesystem set to %s', process_fs_path_config)
        else:
            self.log.debug('The process_fs_path not set. Use default path: /proc')

    def _log_once_per_day(self, message):
        if message in self._partition_error:
            return
        self._partition_error.add(message)
        log.exception(message)

    def check(self, instance):
        """Capture disk stats

        """
        dimensions = self._set_dimensions(None, instance)
        rollup_dimensions = dimensions.copy()

        if instance is not None:
            use_mount = instance.get("use_mount", True)
            send_io_stats = instance.get("send_io_stats", True)
            send_rollup_stats = instance.get("send_rollup_stats", False)
            # If we filter devices, get the list.
            device_blacklist_re = self._get_re_exclusions(instance)
            fs_types_to_ignore = self._get_fs_exclusions(instance)
        else:
            use_mount = True
            send_io_stats = True
            send_rollup_stats = False
            device_blacklist_re = None
            fs_types_to_ignore = set()

        partitions = psutil.disk_partitions(all=True)
        if send_io_stats:
            disk_stats = psutil.disk_io_counters(perdisk=True)
        disk_count = 0
        total_capacity = 0
        total_used = 0
        for partition in partitions:
            if partition.fstype not in fs_types_to_ignore \
                and (not device_blacklist_re or
                     not device_blacklist_re.match(partition.device)):
                try:
                    device_name = self._get_device_name(partition.device)
                    disk_usage = psutil.disk_usage(partition.mountpoint)
                    total_capacity += disk_usage.total
                    total_used += disk_usage.used
                    st = os.statvfs(partition.mountpoint)
                except Exception as ex:
                    exception_name = ex.__class__.__name__
                    self._log_once_per_day('Unable to access partition {} '
                                           'with error: {}'.format(partition,
                                                                   exception_name))
                    continue

                if use_mount:
                    dimensions.update({'mount_point': partition.mountpoint})
                self.gauge("disk.space_used_perc",
                           disk_usage.percent,
                           device_name=device_name,
                           dimensions=dimensions)
                disk_count += 1
                if st.f_files > 0:
                    self.gauge("disk.inode_used_perc",
                               round((float(st.f_files - st.f_ffree) / st.f_files) * 100, 2),
                               device_name=device_name,
                               dimensions=dimensions)
                    disk_count += 1

                log.debug('Collected {0} disk usage metrics for partition {1}'.format(
                    disk_count,
                    partition.mountpoint))
                disk_count = 0
                if send_io_stats:
                    try:
                        stats = disk_stats[device_name]
                        self.rate("io.read_req_sec", round(float(stats.read_count), 2),
                                  device_name=device_name, dimensions=dimensions)
                        self.rate("io.write_req_sec", round(float(stats.write_count), 2),
                                  device_name=device_name, dimensions=dimensions)
                        self.rate("io.read_kbytes_sec",
                                  round(float(stats.read_bytes / 1024), 2),
                                  device_name=device_name, dimensions=dimensions)
                        self.rate("io.write_kbytes_sec",
                                  round(float(stats.write_bytes / 1024), 2),
                                  device_name=device_name, dimensions=dimensions)
                        self.rate("io.read_time_sec", round(float(stats.read_time / 1000), 2),
                                  device_name=device_name, dimensions=dimensions)
                        self.rate("io.write_time_sec", round(float(stats.write_time / 1000), 2),
                                  device_name=device_name, dimensions=dimensions)

                        log.debug('Collected 6 disk I/O metrics for'
                                  'partition {0}'.format(partition.mountpoint))
                    except KeyError:
                        log.debug('No Disk I/O metrics available for'
                                  ' {0}...Skipping'.format(device_name))

        if send_rollup_stats:
            self.gauge("disk.total_space_mb",
                       total_capacity / 1048576,
                       dimensions=rollup_dimensions)
            self.gauge("disk.total_used_space_mb",
                       total_used / 1048576,
                       dimensions=rollup_dimensions)
            log.debug('Collected 2 rolled-up disk usage metrics')

    def _get_re_exclusions(self, instance):
        """Parse device blacklist regular expression"""
        filter = None
        try:
            filter_device_re = instance.get('device_blacklist_re', None)
            if filter_device_re:
                filter = re.compile(filter_device_re)
        except re.error:
            log.error('Error processing regular expression {0}'.format(filter_device_re))

        return filter

    def _get_fs_exclusions(self, instance):
        """parse comma separated file system types to ignore list"""
        file_system_list = set()

        # automatically ignore filesystems not backed by a device
        try:
            for nodevfs in filter(lambda x: x.startswith('nodev\t'), open('/proc/filesystems')):
                file_system_list.add(nodevfs.partition('\t')[2].strip())
        except IOError:
            log.debug('Failed reading /proc/filesystems')

        try:
            file_systems = instance.get('ignore_filesystem_types', None)
            if file_systems:
                # Parse file system types
                file_system_list.update(x.strip() for x in file_systems.split(','))
        except ValueError:
            log.info("Unable to process ignore_filesystem_types.")

        return file_system_list

    def _get_device_name(self, device):
        start = device.rfind("/")
        if start > -1:
            device_name = device[start + 1:]
        else:
            device_name = device

        return device_name



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re

from hacking import core

assert_no_xrange_re = re.compile(r"\s*xrange\s*\(")
assert_True = re.compile(r".*assertEqual\(True, .*\)")
assert_None = re.compile(r".*assertEqual\(None, .*\)")
assert_Not_Equal = re.compile(r".*assertNotEqual\(None, .*\)")
assert_Is_Not = re.compile(r".*assertIsNot\(None, .*\)")
assert_raises_regexp = re.compile(r"assertRaisesRegexp\(")
no_log_warn = re.compile(r".*LOG.warn\(.*\)")
mutable_default_args = re.compile(r"^\s*def .+\((.+=\{\}|.+=\[\])")


@core.flake8ext
def no_mutable_default_args(logical_line):
    msg = "M001: Method's default argument shouldn't be mutable!"
    if mutable_default_args.match(logical_line):
        yield (0, msg)


@core.flake8ext
def no_xrange(logical_line):
    if assert_no_xrange_re.match(logical_line):
        yield (0, "M002: Do not use xrange().")


@core.flake8ext
def validate_assertTrue(logical_line):
    if re.match(assert_True, logical_line):
        msg = ("M003: Unit tests should use assertTrue(value) instead"
               " of using assertEqual(True, value).")
        yield (0, msg)


@core.flake8ext
def validate_assertIsNone(logical_line):
    if re.match(assert_None, logical_line):
        msg = ("M004: Unit tests should use assertIsNone(value) instead"
               " of using assertEqual(None, value).")
        yield (0, msg)


@core.flake8ext
def no_log_warn_check(logical_line):
    if re.match(no_log_warn, logical_line):
        msg = ("M005: LOG.warn is deprecated, please use LOG.warning!")
        yield (0, msg)


@core.flake8ext
def validate_assertIsNotNone(logical_line):
    if re.match(assert_Not_Equal, logical_line) or \
            re.match(assert_Is_Not, logical_line):
        msg = ("M006: Unit tests should use assertIsNotNone(value) instead"
               " of using assertNotEqual(None, value) or"
               " assertIsNot(None, value).")
        yield (0, msg)


@core.flake8ext
def assert_raisesRegexp(logical_line):
    res = assert_raises_regexp.search(logical_line)
    if res:
        yield (0, "M007: assertRaisesRegex must be used instead "
                  "of assertRaisesRegexp")



# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development LP
# Copyright 2016-2017 FUJITSU LIMITED
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
from os import path
import re
import subprocess as sp
from subprocess import CalledProcessError
from subprocess import STDOUT

from oslo_utils import importutils

from monasca_setup import agent_config
from monasca_setup.detection import find_process_cmdline
from monasca_setup.detection import Plugin
from monasca_setup.detection.utils import check_output
from monasca_setup.detection.utils import find_addr_listening_on_port_over_tcp
from monasca_setup.detection import watch_process

log = logging.getLogger(__name__)

_VIA_KAFKA_TOPIC_INDEX = 1
_KAFKA_BIN_LOCATIONS = (
    '',  # refers to kafka binaries being accessible
         # global (i.e. /usr/(local/)?bin)
    '/opt/kafka/bin'
)
_KAFKA_CONSUMER_GROUP_BIN = 'kafka-consumer-groups.sh'
_KAFKA_ZOOKEEPER_SHELL_BIN = 'zookeeper-shell.sh'

_CONSUMER_GROUP_COMMAND_LINE_VALUES_LEN = 7


class Kafka(Plugin):

    """Detect Kafka daemons and sets up configuration to monitor them.
       This plugin configures the kafka_consumer plugin and does not configure any jmx based
       checks against kafka.
       Note this plugin will pull the same information from kafka on each node in the cluster it
       runs on.

       To skip detection consumer groups and topics can be specified with plugin args,
       for example:
       `monasca-setup -d kafka -a "group1=topic1 group2=topic2/topic3"`
       All partitions are assumed for each topic and '/' is used to deliminate more than one
       topic per consumer group.

       For more information see:
           - https://cwiki.apache.org/confluence/display/KAFKA/A+Guide+To+The+Kafka+Protocol
    """

    def __init__(self, template_dir, overwrite=True, args=None, port=9092):
        Plugin.__init__(self, template_dir, overwrite, args)
        self.port = port
        self.zk_url = self._find_zookeeper_url()
        self.config = agent_config.Plugins()

    def _detect(self):
        """Run detection, set self.available True if the service is detected."""
        process_exists = find_process_cmdline('kafka') is not None
        has_dependencies = self.dependencies_installed()

        (self._kafka_consumer_bin,
            self._zookeeper_consumer_bin) = self._find_topic_listing_binaries()

        kafka_has_scripts = (self._kafka_consumer_bin or
                             self._zookeeper_consumer_bin)

        self.available = (process_exists and has_dependencies and
                          kafka_has_scripts)

        if not self.available:
            if not process_exists:
                log.error('Kafka process does not exist.')
            elif not has_dependencies:
                log.error(('Kafka process exists but required '
                           'dependency monasca-common is '
                           'not installed.\n\t'
                           'Please install with: '
                           'pip install monasca-agent[kafka_plugin]'))
            elif not kafka_has_scripts:
                log.error(('Kafka process exists, dependencies are installed '
                           'but neither %s nor %s '
                           'executable was found.'),
                          _KAFKA_CONSUMER_GROUP_BIN,
                          _KAFKA_ZOOKEEPER_SHELL_BIN)

    def _detect_consumers(self):
        """Using zookeeper and a kafka connection find the consumers and associated topics. """
        try:
            log.info("\tInstalling kafka_consumer plugin.")

            kafka_connect_str = self._find_kafka_connection()

            consumers = None

            if self._kafka_consumer_bin:
                consumers = self._detect_consumer_via_kafka()
            if not consumers and self._zookeeper_consumer_bin:
                consumers = self._detect_consumer_via_zookeeper()

            instances = {
                'name': kafka_connect_str,
                'kafka_connect_str': kafka_connect_str,
                'per_partition': False,
                'consumer_groups': consumers or {}
            }
            self.config['kafka_consumer'] = {
                'init_config': None,
                'instances': [instances]
            }

        except Exception as ex:
            log.error('Error Detecting Kafka consumers/topics')
            log.exception(ex)

    def _detect_consumer_via_kafka(self):
        """Detect consumers groups using kafka-consumer-groups"""
        log.info("\tDetecting kafka consumers with {:s} command".format(
            self._kafka_consumer_bin))
        try:
            output = check_output([
                self._kafka_consumer_bin,
                '--zookeeper',
                self.zk_url,
                '--list'
            ], stderr=STDOUT)

            consumers = {}
            consumer_groups = output.splitlines()

            for consumer_group in consumer_groups:
                output = check_output([
                    self._kafka_consumer_bin,
                    '--zookeeper',
                    self.zk_url,
                    '--describe',
                    '--group',
                    consumer_group
                ], stderr=STDOUT)

                lines = output.splitlines()
                topics = {}
                for it, line in enumerate(reversed(lines)):
                    if it == len(lines) - 1 or not line:
                        break
                    values = line.split(',')
                    # There will be always 7 values in output
                    # after splitting the line by ,
                    if (values and len(values) ==
                            _CONSUMER_GROUP_COMMAND_LINE_VALUES_LEN):
                        topics[values[_VIA_KAFKA_TOPIC_INDEX].strip()] = []
                if len(topics.keys()):
                    consumers[consumer_group] = topics

            return consumers
        except Exception as ex:
            log.warn(('Failed to retrieve consumers '
                      'with kafka-consumer-groups.sh. Error is %s'), ex)
        return None

    def _detect_consumer_via_zookeeper(self):
        """Try to get consumers via zookeeper.

        :return: map of consumers group to topics
        :rtype: dict
        """
        log.info("\tDetecting kafka consumers with {:s} command".format(
            self._zookeeper_consumer_bin))
        try:
            consumers = {}
            for consumer in self._ls_zookeeper('/consumers'):
                topics = dict((topic, []) for topic in self._ls_zookeeper(
                    '/consumers/%s/offsets' % consumer))
                if len(topics) > 0:
                    consumers[consumer] = topics
            return consumers
        except Exception as ex:
            log.warn(('Failed to retrieve consumers '
                      'with zookeeper-shell.sh. Error is %s'), ex)

        return None

    def _find_kafka_connection(self):
        listen_ip = find_addr_listening_on_port_over_tcp(self.port)
        if listen_ip:
            log.info("\tKafka found listening on {:s}:{:d}".format(listen_ip, self.port))
        else:
            log.info(
                "\tKafka not found listening on a specific IP (port {:d}),"
                "using 'localhost'".format(
                    self.port))
            listen_ip = 'localhost'

        return "{:s}:{:d}".format(listen_ip, self.port)

    @staticmethod
    def _find_zookeeper_url():
        """Pull the zookeeper url the kafka config.
           :return: Zookeeper url
        """
        zk_connect = re.compile('zookeeper.connect=(.*)')
        try:
            with open('/etc/kafka/server.properties') as settings:
                match = zk_connect.search(settings.read())
        except IOError:
            return None

        if match is None:
            log.error('No zookeeper url found in the kafka server properties.')
            return None

        return match.group(1).split(',')[0]  # Only use the first zk url

    def _ls_zookeeper(self, path):
        """Do a ls on the given zookeeper path.
           I am using the local command line kafka rather than kazoo because it doesn't make
           sense to have kazoo as a dependency only for detection.
        """
        zk_shell = [self._zookeeper_consumer_bin, self.zk_url, 'ls', path]
        try:
            output = check_output(zk_shell, stderr=STDOUT)
        except CalledProcessError:
            log.error('Error running the zookeeper shell to list path %s',
                      path)
            raise

        # The last line is like '[item1, item2, item3]', '[]' or an error message
        # not starting with [
        last_line = output.splitlines()[-1]
        if len(last_line) == 2 or last_line[0] != '[':
            return []
        else:
            return [entry.strip() for entry in last_line.strip('[]').split(',')]

    def _find_topic_listing_binaries(self):

        kafka_bin = zookeper_bin = None

        for location in _KAFKA_BIN_LOCATIONS:
            if not kafka_bin:
                kafka_bin = self._verify_callable_exists(
                    path.join(location, _KAFKA_CONSUMER_GROUP_BIN)
                )
            if not zookeper_bin:
                zookeper_bin = self._verify_callable_exists(
                    path.join(location, _KAFKA_ZOOKEEPER_SHELL_BIN)
                )

        # traversed all locations, this is what we've got
        if kafka_bin:
            log.debug('\tFound %s at %s', _KAFKA_CONSUMER_GROUP_BIN,
                      kafka_bin)
        if zookeper_bin:
            log.debug('\tFound %s at %s', _KAFKA_ZOOKEEPER_SHELL_BIN,
                      zookeper_bin)

        return kafka_bin, zookeper_bin

    @staticmethod
    def _verify_callable_exists(path):
        """Verify if callable exists.

        Method tries to execute binary located
        under path to see if that exists.

        If binary cannot be called, an :py:exc:`OSError` is thrown.
        Effectively that means that binary is not accessible.

        Note:
            If binary is not callable, method returns None

        :param path: path to binary
        :type path: str
        :return: path or None
        :rtype: (str, None)

        """
        try:
            sp.check_output(args=[path], stderr=sp.STDOUT)
        except sp.CalledProcessError:
            log.debug('\tExecutable %s exists', path)
        except OSError:
            log.debug('\tNo executable/file at %s', path)
            return None
        except Exception as ex:
            log.warning('Skipping exception %s', str(ex))
            # note(trebskit) we are not interested in other problems
            # from POV of this method only relevant information
            # is that OSError is not thrown which would be the case
            # if binary is not accessible
            pass
        return path  # return path if if it can be launched

    def build_config(self):
        """Build the config as a Plugins object and return.
            Config includes: consumer_groups (include topics) and kafka_connection_str
        """
        # First watch the process
        self.config.merge(watch_process(['kafka.Kafka'], 'kafka', exact_match=False))
        log.info("\tWatching the kafka process.")

        if not self.dependencies_installed():
            log.warning("Dependencies not installed, skipping Kafka Consumer plugin configuration.")
        elif self.args is not None and len(self.args) > 0:
            kafka_connect_str = self._find_kafka_connection()
            consumers = {}
            service_name = kafka_connect_str
            # Check if the plugin passed in a service name
            # If it did, delete it after use so it doesn't become a consumer group
            if 'service_name' in self.args:
                service_name += '_' + str(self.args.pop('service_name'))
            for key, value in self.args.items():
                value_dict = {topic: [] for topic in value.split('/')}
                consumers[key] = value_dict
            self.config['kafka_consumer'] = {'init_config': None,
                                             'instances': [{'name': service_name,
                                                            'kafka_connect_str': kafka_connect_str,
                                                            'per_partition': False,
                                                            'consumer_groups': consumers}]}
        elif self.zk_url is not None:
            self._detect_consumers()
        return self.config

    def dependencies_installed(self):
        return importutils.try_import('monasca_common', False)



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import re
import subprocess
import sys

import monasca_agent.collector.checks as checks
from monasca_agent.common.psutil_wrapper import psutil
import monasca_agent.common.util as util


log = logging.getLogger(__name__)


class Load(checks.AgentCheck):

    def __init__(self, name, init_config, agent_config):
        super(Load, self).__init__(name, init_config, agent_config)
        self.process_fs_path_config = init_config.get('process_fs_path', None)

    def check(self, instance):
        """Capture load stats

        """

        dimensions = self._set_dimensions(None, instance)

        if util.Platform.is_linux():
            try:
                if self.process_fs_path_config:
                    log.debug(
                        'The path of the process filesystem set to %s',
                        self.process_fs_path_config)
                    loadAvrgProc = open(self.process_fs_path_config + '/loadavg', 'r')
                else:
                    loadAvrgProc = open('/proc/loadavg', 'r')
                uptime = loadAvrgProc.readlines()
                loadAvrgProc.close()
            except Exception:
                log.exception('Cannot extract load using /proc/loadavg')
                return

            uptime = uptime[0]  # readlines() provides a list but we want a string

        elif sys.platform in ('darwin', 'sunos5') or sys.platform.startswith("freebsd"):
            # Get output from uptime
            try:
                uptime = subprocess.Popen(['uptime'],
                                          stdout=subprocess.PIPE,
                                          close_fds=True).communicate()[0]
            except Exception:
                log.exception('Cannot extract load using uptime')
                return

        # Split out the 3 load average values
        load = [res.replace(',', '.') for res in re.findall(r'([0-9]+[\.,]\d+)', uptime)]

        #
        # Normalize the load averages by number of cores
        # so the metric is useful for alarming across
        # hosts with varying core numbers
        #
        num_cores = psutil.cpu_count(logical=True)

        self.gauge('load.avg_1_min',
                   round((float(load[0]) / num_cores), 3),
                   dimensions=dimensions)
        self.gauge('load.avg_5_min',
                   round((float(load[1]) / num_cores), 3),
                   dimensions=dimensions)
        self.gauge('load.avg_15_min',
                   round((float(load[2]) / num_cores), 3),
                   dimensions=dimensions)

        log.debug("Collected 3 load metrics (normalized by {} cores)".format(num_cores))



#!/bin/env python
# Copyright 2017 OrangeLabs
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from copy import deepcopy
import json
import logging
import os
import stat
import subprocess
import time

from monasca_agent.collector.checks import AgentCheck
from monasca_agent.common import keystone
from novaclient import client as nova_client

log = logging.getLogger(__name__)
prerouting_chain = "PREROUTING"
congestion_chain = "congestion"
forward_chain = "FORWARD"

"""Monasca Agent interface for congestion metrics"""


class Congestion(AgentCheck):

    """This Agent provides congestion metrics necessary to monitor network
    performance. It uses ECN marking mechanism to discover the congestion
    in the network. The iptables chains and rules are used to collect ECN
    packets/bytes. Also, the agent provides congestion threshold computed
    from the collected ECN bytes.
    """

    def __init__(self, name, init_config, agent_config, instances=None):
        AgentCheck.__init__(self, name, init_config,
                            agent_config, instances=[{}])
        cache_dir = self.init_config.get('cache_dir')
        self.enable_vm = self.init_config.get('enable_vm')
        self.enable_ecn = self.init_config.get('enable_ecn')
        self.s_factor = self.init_config.get('s_factor')
        self.collect_period = self.init_config.get('collect_period')
        self.cong_cache_file = os.path.join(cache_dir,
                                            'congestion_status.json')
        self.session = keystone.get_session(**self.init_config)
        self.chain_exist = False
        self.rule_exist = False
        self._check_chain()
        self.checked = []
        if self.enable_ecn:
            self._activate_ecn()

    def check(self, instance):
        """Extend check method to collect and update congestion metrics.
        """
        dimensions = self._set_dimensions({'service': 'networking',
                                           'component': 'neutron'}, instance)
        self.sample_time = float("{:9f}".format(time.time()))
        """Check iptables information and verify/install the ECN rule for
        specific hypervisor"""
        ip_list = self._get_hypervisors()
        """update congestion metrics for each remote hypervisor"""
        for name, ip in ip_list.items():
            if name != self.hostname and name not in self.checked:
                self.checked.append(name)
                dimensions.update({'hostname': name})
                self._update_metrics(name, ip, dimensions)
                """update congestion metrics for vms if this option
                was enabled"""
                if self.enable_vm:
                    ip_vm_list = self._get_vms(name)
                    if ip_vm_list:
                        for name_vm, ip_vm in ip_vm_list.items():
                            if name_vm not in self.checked:
                                self.checked.append(name_vm)
                                dimensions.update({'device': name_vm})
                                self._update_metrics(name_vm, ip_vm,
                                                     dimensions)

    def _update_metrics(self, name, ip, dimensions):
        """This method updates congestion metrics and cache and sends
        them to monasca API for further treatment or evaluation.
        """
        cong_cache = self._load_cong_cache()
        rule_data = self._get_counters(ip, congestion_chain)
        if not rule_data:
            match = "tos --tos 0x03"
            action = "MARK --set-mark 3"
            self._add_rule(congestion_chain, ip, match, action)
            rule_data = self._get_counters(ip, congestion_chain)
        if name not in cong_cache:
            cong_cache[name] = {}
            """initalize cache values"""
            cong_cache[name]['ecn.cong.rate'] = {'value': 0}
            cong_cache[name]['ecn.bytes'] = {'value': 0}
            cong_cache[name]['ecn_bytes_sum'] = {'value': 0}
            cong_cache[name]['ecn.packets'] = {'value': 0}
            cong_cache[name]['ecn.packets_sum'] = {'value': 0}
        ecn_packets = int(rule_data[0]) - \
            cong_cache[name]['ecn.packets_sum']['value']
        cong_cache[name]['ecn.packets_sum']['value'] = int(rule_data[0])
        ecn_bytes = int(rule_data[1]) - \
            cong_cache[name]['ecn_bytes_sum']['value']
        cong_cache[name]['ecn_bytes_sum']['value'] = int(rule_data[1])
        """ecn congestion average equation"""
        ecn_cong_avg = self.s_factor * \
            (ecn_bytes * 8 / 1000 / self.collect_period)
        ecn_cong_rate_prev = cong_cache[name]['ecn.cong.rate']['value']
        """Actual ecn congestion rate takes into consideration the previous
        value with a certain percentage. The result is expressed in kbps"""
        ecn_cong_rate = ecn_cong_avg + (1 - self.s_factor) * ecn_cong_rate_prev
        """Update the cache file with new metric values"""
        cong_cache[name]['ecn.packets'] = {'timestamp': self.sample_time,
                                           'value': ecn_packets}
        cong_cache[name]['ecn.bytes'] = {'timestamp': self.sample_time,
                                         'value': ecn_bytes}
        cong_cache[name]['ecn.cong.rate'] = {'timestamp': self.sample_time,
                                             'value': ecn_cong_rate}
        self.log.info("metric updated for %s.", name)
        self.gauge('ecn.packets', ecn_packets, dimensions)
        self.gauge('ecn.bytes', ecn_bytes, dimensions)
        self.gauge('ecn.cong.rate', ecn_cong_rate, dimensions)
        self._update_cong_cache(cong_cache)

    def _check_chain(self):
        """Verify if the necessary iptables' chains are in place
        """
        for chain in [congestion_chain, prerouting_chain]:
            self._get_rule(chain)
        """Add new congestion chain if it's not in the table"""
        if not self.chain_exist:
            self._add_chain(congestion_chain)
        """Redirect any packet received by Prerouting chain to congestion
        chain"""
        if not self.rule_exist:
            self._add_rule(prerouting_chain, None, None, congestion_chain)

    def _activate_ecn(self):
        """Ensures that the ECN marking is enable on each tap interface
        """
        tos_rule = "TOS --set-tos 0x02/0xff"
        taps = None
        """Collect tap intefaces attached to linux bridge"""
        try:
            taps = subprocess.check_output(
                "brctl show | awk '{print $1}' | grep tap",
                shell=True, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            self.log.error(e.output)
        if taps:
            taps = filter(None, taps.split('\n'))
            """Collect installed rules in Forward chain"""
            forw_rules = self._get_rule(forward_chain)
            for tap in taps:
                tap = tap + " --physdev-is-bridged"
                if not self._find_tap(tap, forw_rules, tos_rule):
                    """Enable ECN"""
                    match = "physdev --physdev-in " + tap
                    self._add_rule(forward_chain, None, match, tos_rule)
                    self.log.info("ECN is enabled for %s interface.", tap)

    def _find_tap(self, tap, chain, tos_rule):
        for rule in chain:
            """Check if the rule was applied to tap interface"""
            if (tap + " -j " + tos_rule) in rule:
                return True
        return False

    def _add_chain(self, chain):
        """This method adds 'chain' into iptables.
        """
        command = "sudo iptables -t mangle -N " + chain + " -w"
        try:
            subprocess.check_output(
                command, shell=True, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            self.log.error(
                "Command {} return with error (code {}): {}"
                    .format(e.cmd, e.returncode, e.output))
        self.log.info("New %s chain was added to mangle table.", chain)

    def _add_rule(self, chain, ip, match, action):
        """Add new iptables rule based on the given arguments.
        """
        basic_rule = "sudo iptables -t mangle -A "
        if chain == prerouting_chain:
            command = basic_rule + chain + " -j " + action + " -w "
        if chain == congestion_chain:
            command = basic_rule + chain + " -s " + ip + " -m " + match + \
                " -j " + action + " -w"
        if chain == forward_chain:
            command = basic_rule + chain + " -m " + match + \
                " -j " + action + " -w"
        try:
            subprocess.check_output(
                command, shell=True, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            self.log.error(
                "Command {} return with error (code {}): {}"
                    .format(e.cmd, e.returncode, e.output))
        self.log.info("New rule was added to %s ", chain)

    def _get_rule(self, chain):
        """Search in the iptables chains if 'chain' exist.
        """
        command = "sudo iptables -S -t mangle -w"
        forw_rules = []
        try:
            output = subprocess.check_output(
                command, shell=True, stderr=subprocess.STDOUT,
                universal_newlines=True)
        except subprocess.CalledProcessError as e:
            self.log.error(
                "Command {} return with error (code {}): {}"
                    .format(e.cmd, e.returncode, e.output))
        output = filter(None, output.split('\n'))
        if chain == congestion_chain:
            for rule in output:
                if '-N congestion' in rule:
                    self.chain_exist = True
                    break
        if chain == prerouting_chain:
            for rule in output:
                if '-A PREROUTING -j congestion' in rule:
                    self.rule_exist = True
                    break
        if chain == forward_chain:
            for rule in output:
                if '-A FORWARD' in rule:
                    forw_rules.append(rule)
            return forw_rules
        return 0

    def _get_hypervisors(self):
        """Connect to nova client and get the name/ip of all remote compute
        (hypervisor).
        """
        hyp_list = {}
        n_client = self._get_nova_client()
        hypervisors = n_client.hypervisors.list()
        for hypervisor in hypervisors:
            name = hypervisor.__dict__['hypervisor_hostname']
            ip = hypervisor.__dict__['host_ip']
            if name not in hyp_list:
                hyp_list[name] = ip
        return hyp_list

    def _get_vms(self, compute_name):
        """Connect to nova client and collect the name and the ip of all VMs
        deployed in a specific compute_name.
        """
        vm_list = {}
        n_client = self._get_nova_client()
        try:
            instances = n_client.servers.list(
                search_opts={'all_tenants': 1, 'host': compute_name})
        except Exception as e:
            self.log.error(
                "%s : No instances hosted in %s compute. ", e, compute_name)
            vm_list = {}
        if instances:
            for instance in instances:
                inst_name = instance.__getattr__('name')
                for net in instance.addresses:
                    for ip in instance.addresses[net]:
                        if (ip['OS-EXT-IPS:type'] == 'fixed' and
                                ip['version'] == 4):
                            vm_list[inst_name] = ip['addr']
        return vm_list

    def _get_counters(self, ip, chain):
        """Collect packets and bytes of each source 'ip' existing in 'chain'.
        """
        counters = ()
        command = "sudo iptables -L " + chain + " -v -t mangle -w"
        try:
            output = subprocess.check_output(
                command, shell=True, stderr=subprocess.STDOUT,
                universal_newlines=True)
        except subprocess.CalledProcessError as e:
            self.log.error(
                "Command {} return with error (code {}): {}"
                    .format(e.cmd, e.returncode, e.output))
        output = filter(None, output.split('\n'))
        for line in output:
            if 'tos match0x03' in line:
                line = filter(None, line.split(' '))
                if str(ip) == line[7]:
                    packet = self._convert_data(line[0])
                    bytes = self._convert_data(line[1])
                    counters = (packet, bytes)
        return counters

    def _convert_data(self, data):
        """Convert any string that contains a K or M letter to an integer.
        """
        if 'K' in str(data):
            data = int(data.replace('K', '')) * 1000
        if 'M' in str(data):
            data = int(data.replace('M', '')) * 1000000
        return data

    def _load_cong_cache(self):
        """Load congestion metrics from the previous measurement stored as
        cache file in the hard disk.
        """
        load_cong_cache = {}
        try:
            with open(self.cong_cache_file, 'r') as cache_json:
                load_cong_cache = json.load(cache_json)
        except (IOError, TypeError, ValueError):
            self.log.warning(
                "Congestion cache missing or corrupt, rebuilding.")
            load_cong_cache = {}
        return load_cong_cache

    def _update_cong_cache(self, cong_cache):
        """update cache file."""
        write_cong_cache = deepcopy(cong_cache)
        try:
            with open(self.cong_cache_file, 'w') as cache_json:
                json.dump(write_cong_cache, cache_json)
            if stat.S_IMODE(os.stat(self.cong_cache_file).st_mode) != 0o600:
                os.chmod(self.cong_cache_file, 0o600)
            self.log.warning("Your cache file is updated : %s ", time.time())
        except IOError as e:
            self.log.error("Cannot write to {0}: {1}".
                           format(self.cong_cache_file, e))

    def _get_nova_client(self):
        """Get nova client object based on configured parameters.
        """
        region_name = self.init_config.get('region_name')
        endpoint_type = self.init_config.get("endpoint_type", "publicURL")
        nc = nova_client.Client(2, session=self.session,
                                endpoint_type=endpoint_type,
                                service_type="compute",
                                region_name=region_name)

        return nc



# (C) Copyright 2015-2016,2018 Hewlett Packard Enterprise Development LP
# Copyright 2017 Fujitsu LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import copy
import json
import logging

from keystoneauth1.exceptions import base as keystoneauth_exception
from monascaclient import client
from osc_lib import exceptions

from monasca_agent.common import keystone

log = logging.getLogger(__name__)


class MonascaAPI(object):

    """Sends measurements to MonascaAPI
        Any errors should raise an exception so the transaction calling
        this is not committed
    """

    LOG_INTERVAL = 10  # messages

    def __init__(self, config):
        """Initialize Mon api client connection."""
        self._config = config
        self._mon_client = None
        self._api_version = '2_0'

        self._failure_reason = None
        self._log_interval_remaining = 1
        self._current_number_measurements = 0
        self._max_buffer_size = int(config['max_buffer_size'])
        self._max_batch_size = int(config['max_batch_size'])
        self._max_measurement_buffer_size = int(
            config['max_measurement_buffer_size'])

        if self._max_buffer_size > -1:
            log.debug("'max_buffer_size' is deprecated. Please use"
                      " 'max_measurement_buffer_size' instead")
            if self._max_measurement_buffer_size > -1:
                log.debug("Overriding 'max_buffer_size' option with new"
                          " 'max_measurment_buffer_size' option")
                self._max_buffer_size = -1

        self.backlog_send_rate = int(config['backlog_send_rate'])
        if self._max_buffer_size > -1:
            self.message_queue = collections.deque(maxlen=self._max_buffer_size)
        else:
            self.message_queue = collections.deque()
        self.write_timeout = int(config['write_timeout'])
        # 'amplifier' is completely optional and may not exist in the config
        try:
            self.amplifier = int(config['amplifier'])
        except KeyError:
            self.amplifier = None

    def _post(self, measurements, tenant=None):
        """Does the actual http post
            measurements is a list of Measurement
        """
        kwargs = {
            'jsonbody': measurements
        }

        if tenant:
            kwargs['tenant_id'] = tenant

        if not self._mon_client:
            self._mon_client = self._get_mon_client()
            if not self._mon_client:
                # Keystone is down, queue the message
                self._queue_message(kwargs.copy(), "Keystone API is down or unreachable")
                return

        if self._send_message(**kwargs):
            if len(self.message_queue) > 0:
                messages_sent = 0
                for index in range(0, len(self.message_queue)):
                    if index < self.backlog_send_rate:

                        msg = json.loads(self.message_queue.pop())

                        if self._send_message(**msg):
                            messages_sent += 1
                            for value in msg.values():
                                self._current_number_measurements -= len(value)
                        else:
                            self._queue_message(msg, self._failure_reason)
                            break
                    else:
                        break
                log.info("Sent {0} messages from the backlog.".format(messages_sent))
                log.info("{0} messages remaining in the queue.".format(len(self.message_queue)))
                self._log_interval_remaining = 0
        else:
            self._queue_message(kwargs.copy(), self._failure_reason)

    def post_metrics(self, measurements):
        """post_metrics
            given [Measurement, ...], format the request and post to
            the monitoring api
        """
        # Add default dimensions
        for envelope in measurements:
            measurement = envelope['measurement']
            if isinstance(measurement['dimensions'], list):
                measurement['dimensions'] = dict([(d[0], d[1]) for d in measurement['dimensions']])

        # Split out separate POSTs for each delegated tenant (includes 'None')
        tenant_group = {}
        for envelope in measurements:
            measurement = envelope['measurement']
            tenant = envelope['tenant_id']
            tenant_group.setdefault(tenant, []).append(copy.deepcopy(measurement))
            if self._max_batch_size and len(tenant_group[tenant]) >= self._max_batch_size:
                self._post(tenant_group[tenant], tenant)
                del tenant_group[tenant]

        for tenant in tenant_group:
            self._post(tenant_group[tenant], tenant)

    def _get_mon_client(self):
        try:
            k = keystone.Keystone(self._config)
            endpoint = k.get_monasca_url()
            session = k.get_session()
            c = client.Client(
                api_version=self._api_version,
                endpoint=endpoint,
                session=session,
                timeout=self.write_timeout,
                **keystone.get_args(self._config)
            )
            return c
        except keystoneauth_exception.ClientException as ex:
            log.error('Failed to initialize Monasca client. {}'.format(ex))

    def _send_message(self, **kwargs):
        try:
            self._mon_client.metrics.create(**kwargs)
            return True
        except exceptions.ClientException as ex:
            log.exception("ClientException: error sending "
                          "message to monasca-api.")
            self._failure_reason = ('Error sending message to '
                                    'the Monasca API: {0}').format(str(ex))
        except Exception:
            log.exception("Error sending message to Monasca API.")
            self._failure_reason = 'The Monasca API is DOWN or unreachable'

        return False

    def _queue_message(self, msg, reason):
        if self._max_buffer_size == 0 or self._max_measurement_buffer_size == 0:
            return

        self.message_queue.append(json.dumps(msg))

        for value in msg.values():
            self._current_number_measurements += len(value)

        if self._max_measurement_buffer_size > -1:
            while self._current_number_measurements > self._max_measurement_buffer_size:
                self._remove_oldest_from_queue()

        if self._log_interval_remaining <= 1:
            log.warn("{0}. Queuing the messages to send later...".format(reason))
            log.info("Current agent queue size: {0} of {1}.".format(len(self.message_queue),
                                                                    self._max_buffer_size))
            log.info("Current measurements in queue: {0} of {1}".format(
                self._current_number_measurements, self._max_measurement_buffer_size))

            log.info(
                "A message will be logged for every {0} messages queued.".format(
                    MonascaAPI.LOG_INTERVAL))
            self._log_interval_remaining = MonascaAPI.LOG_INTERVAL
        else:
            self._log_interval_remaining -= 1

    def _remove_oldest_from_queue(self):
        removed_batch = json.loads(self.message_queue.popleft())
        num_discarded = 0
        for value in removed_batch.values():
            num_discarded += len(value)
        self._current_number_measurements -= num_discarded
        log.warn("Queue too large, discarding oldest batch: {0} measurements discarded".format(
            num_discarded))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import re

import monasca_agent.collector.checks as checks

_LXC_CGROUP_PWD = '/sys/fs/cgroup'  # nosec B105
_LXC_CGROUP_CPU_PWD = '{0}/cpu/lxc'.format(_LXC_CGROUP_PWD)
_LXC_CGROUP_CPUSET_PWD = '{0}/cpuset/lxc'.format(_LXC_CGROUP_PWD)
_LXC_CGROUP_MEM_PWD = '{0}/memory/lxc'.format(_LXC_CGROUP_PWD)
_LXC_CGROUP_DISK_PWD = '{0}/blkio/lxc'.format(_LXC_CGROUP_PWD)

_LXC_NET_REGEX = re.compile(r'(\w+):(.+)')
_LXC_DISK_REGEX = re.compile(r'(\w+)\s(\d+)')


class LXC(checks.AgentCheck):
    """Agent to collect LXC cgroup information.

    The information is mostly based on cgroup files of each container.
    """

    def check(self, instance):
        self.instance = instance
        self.containers = self._containers_name()
        self.increment('running_containers', len(self.containers),
                       {'service': 'lxc'})
        for container_name in self.containers:
            self._collect_cpu_metrics(container_name)
            self._collect_mem_metrics(container_name)
            self._collect_swap_metrics(container_name)
            self._collect_net_metrics(container_name)
            self._collect_disk_metrics(container_name)

    def _containers_name(self):
        container_name = self.instance.get('container')
        if container_name == 'all':
            return [name for name in os.listdir(_LXC_CGROUP_CPU_PWD)
                    if os.path.isdir('{0}/{1}'.format(_LXC_CGROUP_CPU_PWD,
                                                      name))]

        if os.path.isdir('{0}/{1}'.format(_LXC_CGROUP_CPU_PWD,
                                          container_name)):
            self.log.info('\tContainer name: ' + container_name)
            return [container_name]
        else:
            self.log.error('\tContainer {0} was not found'
                           .format(container_name))
            return

    def _collect_cpu_metrics(self, container_name):
        if not self.instance.get('cpu', True):
            return
        metrics = self._get_cpu_metrics(container_name)
        cpu_dimensions = self._get_dimensions(container_name)
        for metric, value in metrics.items():
            self.gauge(metric, value, dimensions=cpu_dimensions)

    def _collect_mem_metrics(self, container_name):
        if not self.instance.get('mem', True):
            return
        metrics = self._get_mem_metrics(container_name)
        mem_dimensions = self._get_dimensions(container_name)
        for metric, value in metrics.items():
            self.gauge(metric, value, dimensions=mem_dimensions)

    def _collect_swap_metrics(self, container_name):
        if not self.instance.get('swap', True):
            return
        metrics = self._get_swap_metrics(container_name)
        if metrics:
            swap_dimensions = self._get_dimensions(container_name)
            for metric, value in metrics.items():
                self.gauge(metric, value, dimensions=swap_dimensions)

    def _collect_net_metrics(self, container_name):
        if not self.instance.get('net', True):
            return
        metrics = self._get_net_metrics(container_name)
        for iface_name, iface_metrics in metrics.items():
            net_dimensions = self._get_dimensions(container_name,
                                                  {'iface': iface_name})
            for metric, value in iface_metrics.items():
                self.gauge(metric, value, dimensions=net_dimensions)

    def _collect_disk_metrics(self, container_name):
        if not self.instance.get('blkio', True):
            return
        metrics = self._get_disk_metrics(container_name)
        disk_dimensions = self._get_dimensions(container_name)
        for metric, value in metrics.items():
            self.gauge(metric, value, dimensions=disk_dimensions)

    def _get_cpu_metrics(self, container_name):
        """Get metrics from cpuacct.usage cgroup file

            :return: a dictionary containing cpu metrics defined on container
            cgroup
        """
        metrics = {}
        cpu_cgroup = '{0}/{1}/'.format(_LXC_CGROUP_CPU_PWD, container_name)
        metrics['cpuacct.usage'] = int(open(cpu_cgroup + 'cpuacct.usage', 'r')
                                       .readline().rstrip('\n'))
        cpuacct_usage_percpu = open(cpu_cgroup + 'cpuacct.usage_percpu', 'r')\
            .readline().rstrip(' \n').split(' ')
        for cpu in range(len(cpuacct_usage_percpu)):
            metrics['cpuacct.usage_percpu.cpu{0}'.format(cpu)] = \
                int(cpuacct_usage_percpu[cpu])
        metrics_stat = self._get_metrics_by_file(cpu_cgroup + 'cpuacct.stat',
                                                 'cpuacct')
        metrics.update(metrics_stat)
        return metrics

    def _get_mem_metrics(self, container_name):
        """Get metrics from memory.stat and memory cgroup file

           :returns: a dictionary containing memory metrics defined on
           container cgroup
        """
        mem_cgroup = '{0}/{1}/'.format(_LXC_CGROUP_MEM_PWD, container_name)
        metrics = self._get_metrics_by_file(mem_cgroup + 'memory.stat',
                                            'memory')
        # Get others cgroup memory values
        with open('{0}/memory.usage_in_bytes'.format(mem_cgroup)) as mem_file:
            metrics['memory.usage_in_bytes'] = int(mem_file.read())
        return metrics

    def _get_swap_metrics(self, container_name):
        """Get swap metrics from memory cgroup file.

           If swapaccount is defined, you can control swap memory. To active swap
           memory control, set GRUB_CMDLINE_LINUX_DEFAULT="swapaccount=1" on
           /etc/default/grub file and restart the machine.

           :returns: a dictionary containing swap metrics defined on
           container cgroup
        """
        metrics = {}
        swap_file = '{0}/{1}/memory.memsw.usage_in_bytes'.format(_LXC_CGROUP_MEM_PWD,
                                                                 container_name)
        if os.path.isfile(swap_file):
            with open(swap_file) as mem_file:
                metrics['memory.memsw.usage_in_bytes'] = int(mem_file.read())
        else:
            self.log.error('Swap cgroup fine not found. '
                           'Verify if swapaccount control is enable')
        return metrics

    def _get_net_metrics(self, container_name):
        """Get metrics for each net interface found

        :returns: a dictionary containing metrics regarding each
        net interface found, in the format:
        { 'lo': { 'net.rx.bytes': 1234 }, ...}
        """
        metrics = {}
        pid = self._get_pid_container(container_name)
        net_cgroup = '/proc/{0}/net/'.format(pid)
        with open(net_cgroup + 'dev', 'r') as dev_file:
            for line in dev_file:
                iface = re.search(_LXC_NET_REGEX, line)
                if iface:
                    iface_name = iface.group(1)
                    iface_info = iface.group(2).split()
                    metrics[iface_name] = {
                        'net.rx.bytes': int(iface_info[0]),
                        'net.rx.packets': int(iface_info[1]),
                        'net.rx.errs': int(iface_info[2]),
                        'net.rx.drop': int(iface_info[3]),
                        'net.rx.fifo': int(iface_info[4]),
                        'net.rx.frame': int(iface_info[5]),
                        'net.rx.compressed': int(iface_info[6]),
                        'net.rx.multicast': int(iface_info[7]),
                        'net.tx.bytes': int(iface_info[8]),
                        'net.tx.packets': int(iface_info[9]),
                        'net.tx.errs': int(iface_info[10]),
                        'net.tx.drop': int(iface_info[11]),
                        'net.tx.fifo': int(iface_info[12]),
                        'net.tx.frame': int(iface_info[13]),
                        'net.tx.compressed': int(iface_info[14]),
                        'net.tx.multicast': int(iface_info[15])
                    }
        return metrics

    def _get_disk_metrics(self, container_name):
        """Get metrics blkio.throttle.io_service_bytes from cgroup file

            :return: a dictionary containing blkio metrics used to verify disk
            cgroup usage
        """
        metrics = {}
        disk_cgroup = '{0}/{1}/blkio.throttle.io_service_bytes'.format(
                      _LXC_CGROUP_DISK_PWD, container_name)
        with open(disk_cgroup, 'r') as disk_file:
            for line in disk_file:
                disk = re.search(_LXC_DISK_REGEX, line)
                if disk:
                    disk_key = 'blkio.{0}'.format(disk.group(1)).lower()
                    disk_value = disk.group(2)
                    metrics[disk_key] = int(disk_value)
        return metrics

    def _get_metrics_by_file(self, filename, pre_key):
        """Some cgroup files have a pattern 'key value' that can be easily
        handled to a dictionary
        """
        metrics = {}
        with open(filename, 'r') as cgroup_file:
            for line in cgroup_file:
                resource_post_key, resource_value = line.split(' ')
                resource_key = '{0}.{1}'.format(pre_key, resource_post_key)
                metrics[resource_key] = int(resource_value)
        return metrics

    def _get_dimensions(self, container_name, options=None):
        dimensions = {'container_name': container_name,
                      'service': 'lxc'}
        if options:
            dimensions.update(options)
        return self._set_dimensions(dimensions, self.instance)

    def _get_pid_container(self, container_name):
        cpu_tasks = '{0}/{1}/tasks'.format(_LXC_CGROUP_CPU_PWD,
                                           container_name)
        pid = open(cpu_tasks, 'r').readline().rstrip('\n')
        return pid



# Copyright 2017 OrangeLabs
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import configparser
import logging
import os

from monasca_agent.common.psutil_wrapper import psutil
import monasca_setup.agent_config
import monasca_setup.detection

log = logging.getLogger(__name__)

# Directory to use for metric caches
cache_dir = "/dev/shm"
# Enable vm congestion check
enable_vm = True
# Ensure that ECN marking is enabled
enable_ecn = True
# Smoothing factor used to compute ecn congestion rate
s_factor = 0.1
# Period of time in second to collect metrics
collect_period = 30
# Acceptable arguments
acceptable_args = ['username', 'password', 'project_name',
                   'auth_url', 'cache_dir', 'enable_vm', 'enable_ecn',
                   'region_name', 's_factor', 'collect_period']


class Congestion(monasca_setup.detection.Plugin):

    """Configures congestion detection plugin."""

    def _detect(self):
        """Run detection, set self.available True if the service is
           detected.
        """
        self.available = False
        # Start with plugin-specific configuration parameters
        # Try to detect the location of the Nova configuration file.
        # Walk through the list of processes, searching for 'nova-compute'
        # process with 'nova.conf' inside one of the parameters
        nova_conf = None
        for proc in psutil.process_iter():
            try:
                cmd = proc.as_dict(['cmdline'])['cmdline']
                if (len(cmd) > 2 and 'python' in cmd[0] and
                        'nova-compute' in cmd[1]):
                    conf_indexes = [cmd.index(y)
                                    for y in cmd if 'nova.conf' in y]
                    if not conf_indexes:
                        continue
                    param = conf_indexes[0]
                    if '=' in cmd[param]:
                        nova_conf = cmd[param].split('=')[1]
                    else:
                        nova_conf = cmd[param]
            except (IOError, psutil.NoSuchProcess):
                # Process has already terminated, ignore
                continue
        if (nova_conf is not None and os.path.isfile(nova_conf)):
            self.available = True
            self.nova_conf = nova_conf

    def build_config(self):
        """Build the config as a Plugins object and return.  """
        config = monasca_setup.agent_config.Plugins()
        log.info("Configuring congestion plugin")
        nova_cfg = configparser.SafeConfigParser()
        log.info("\tUsing nova configuration file {0}".format(self.nova_conf))
        nova_cfg.read(self.nova_conf)
        # Which configuration options are needed for the plugin YAML?
        # Use a dict so that they can be renamed later if necessary
        cfg_needed = {
            'username': 'username', 'password': 'password',
            'project_name': 'project_name'}
        cfg_section = 'keystone_authtoken'
        region_name_sec = 'neutron'
        init_config = {
            'cache_dir': cache_dir,
            'enable_vm': enable_vm,
            'enable_ecn': enable_ecn,
            's_factor': s_factor,
            'collect_period': collect_period}
        for option in cfg_needed:
            init_config[option] = nova_cfg.get(
                cfg_section, cfg_needed[option])
        init_config['region_name'] = nova_cfg.get(
            region_name_sec, 'region_name')
        # Create an identity URI (again, slightly different for Devstack)
        if nova_cfg.has_option(cfg_section, 'auth_url'):
            init_config['auth_url'] = nova_cfg.get(cfg_section, 'auth_url')
        else:
            init_config['auth_url'] = nova_cfg.get(
                cfg_section, 'identity_uri')

        config = monasca_setup.agent_config.Plugins()
        config['congestion'] = {
            'init_config': init_config, 'instances': []}
        return config

    def dependencies_installed(self):
        return True



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import monasca_setup.detection


class Designate(monasca_setup.detection.ServicePlugin):

    """Detect Designate daemons and setup configuration to monitor them.

    """

    def __init__(self, template_dir, overwrite=True, args=None):
        service_params = {
            'args': args,
            'template_dir': template_dir,
            'overwrite': overwrite,
            'service_name': 'dns',
            'process_names': ['designate-api',
                              'designate-central',
                              'designate-mdns',
                              'designate-pool-manager',
                              'designate-zone-manager',
                              'designate-sink',
                              'designate-agent'],
            'service_api_url': 'http://localhost:9001',
            'search_pattern': '.*200 OK.*',
        }

        super(Designate, self).__init__(service_params)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
import unittest
from collections import defaultdict
from tempfile import mkdtemp
import shutil
import os
from array import array

from monasca_agent.collector.checks_d import swift_handoffs


class FakeLogger(object):
    def __init__(self):
        self.log = {'error': [],
                    'warning': []}

    def _write_msg(self, msg, key):
        self.log[key].append(msg)

    def error(self, msg):
        self._write_msg(msg, 'error')

    def warning(self, msg):
        self._write_msg(msg, 'warning')

    def get_loglines(self, key):
        return self.log[key]


class MockSwiftHandoffs(swift_handoffs.SwiftHandoffs):
    def __init__(self):
        swift_handoffs.swift_loaded = True
        super(MockSwiftHandoffs, self).__init__(
            name='swift_handoffs',
            init_config={},
            instances=[],
            agent_config={}
        )
        self.log = FakeLogger()
        self.reset_gauge()

    def gauge(self, key, value, dimensions, *args, **kwargs):
        self.gauge_called = True
        self.gauge_calls[key].append(value)
        for k, v in dimensions.items():
            self.dimensions[k].add(v)

    def reset_gauge(self):
        self.gauge_called = False
        self.gauge_calls = defaultdict(list)
        self.dimensions = defaultdict(set)


class MockRing(object):
    def __init__(self, *args):
        self.devs = [
            {u'device': u'sdb1', u'id': 0, u'ip': u'127.0.0.1',
             u'meta': u'', u'port': 6010, u'region': 1,
             u'replication_ip': u'127.0.0.1', u'replication_port': 6010,
             u'weight': 1.0, u'zone': 1},
            {u'device': u'sdb2', u'id': 1, u'ip': u'127.0.0.1',
             u'meta': u'', u'port': 6010, u'region': 1,
             u'replication_ip': u'127.0.0.1', u'replication_port': 6010,
             u'weight': 1.0, u'zone': 1},
            {u'device': u'sdb3', u'id': 2, u'ip': u'127.0.0.2',
             u'meta': u'', u'port': 6010, u'region': 1,
             u'replication_ip': u'127.0.0.2', u'replication_port': 6010,
             u'weight': 1.0, u'zone': 1},
            {u'device': u'sdb4', u'id': 3, u'ip': u'127.0.0.2',
             u'meta': u'', u'port': 6010, u'region': 1,
             u'replication_ip': u'127.0.0.2', u'replication_port': 6010,
             u'weight': 1.0, u'zone': 1}]

        self._replica2part2dev_id = [
            array('H', [3, 0, 2, 1, 2, 3, 0, 1, 3, 3, 0, 1, 2, 1, 0, 2]),
            array('H', [0, 2, 1, 3, 1, 0, 2, 3, 0, 0, 2, 3, 1, 3, 2, 1]),
            array('H', [2, 1, 3, 0, 3, 2, 1, 0, 2, 2, 1, 0, 3, 0, 1, 3])]


class SwiftHandoffsTest(unittest.TestCase):
    def setUp(self):
        super(SwiftHandoffsTest, self).setUp()
        self.swift_handoffs = MockSwiftHandoffs()
        self.tmpdir = mkdtemp()
        self.datadir = os.path.join(self.tmpdir, 'datadir')
        self.ring = os.path.join(self.tmpdir, 'object.ring.gz')
        os.mkdir(self.datadir)
        os.mknod(self.ring)
        self.expected_dev2part = {
            u'sdb1': {0, 1, 3, 5, 6, 7, 8, 9, 10, 11, 13, 14},
            u'sdb2': {1, 2, 3, 4, 6, 7, 10, 11, 12, 13, 14, 15},
            u'sdb3': {0, 1, 2, 4, 5, 6, 8, 9, 10, 12, 14, 15},
            u'sdb4': {0, 2, 3, 4, 5, 7, 8, 9, 11, 12, 13, 15}}

        self.expected_handoffs = {
            u'sdb1': {2, 4, 12, 15},
            u'sdb2': {0, 5, 8, 9},
            u'sdb3': {3, 7, 11, 13},
            u'sdb4': {1, 6, 10, 14}}

    def tearDown(self):
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    @mock.patch('monasca_agent.collector.checks_d.swift_handoffs.Ring',
                MockRing)
    def test_get_ring_and_datadir(self):
        def do_test(path, expected_ringname, expected_datadir):
            _ring, name, datadir = swift_handoffs.get_ring_and_datadir(path)
            self.assertEqual(name, expected_ringname)
            self.assertEqual(datadir, expected_datadir)

        for prefix in ('/etc/swift/{}', './{}', 'some/other/loc/{}'):
            test_cases = (
                (prefix.format('object.ring.gz'), 'object', 'objects'),
                (prefix.format('object-1.ring.gz'), 'object-1', 'objects-1'),
                (prefix.format('object-2.ring.gz'), 'object-2', 'objects-2'),
                (prefix.format('object-50.ring.gz'), 'object-50', 'objects-50'),
                (prefix.format('container.ring.gz'), 'container', 'containers'),
                (prefix.format('account.ring.gz'), 'account', 'accounts'))
            for path, ex_ringname, ex_datadir in test_cases:
                do_test(path, ex_ringname, ex_datadir)

    def test_check_missing_options(self):
        # missing device (path to devices mount point), and default doesn't
        # exist
        instance = {'ring': self.ring}
        with mock.patch('os.path.exists', return_value=False):
            self.swift_handoffs.check(instance)
        self.assertIn('devices must exist or be a directory',
                      self.swift_handoffs.log.get_loglines('error'))
        self.swift_handoffs.log = FakeLogger()

        # a device that isn't a dir
        instance = {'ring': self.ring,
                    'devices': '{}/random'.format(self.datadir)}
        with mock.patch('os.path.exists', return_value=True), \
                mock.patch('os.path.isdir', return_value=False):
            self.swift_handoffs.check(instance)
        self.assertIn('devices must exist or be a directory',
                      self.swift_handoffs.log.get_loglines('error'))
        self.swift_handoffs.log = FakeLogger()

        # missing ring
        instance = {'devices': self.datadir}
        self.swift_handoffs.check(instance)
        self.assertIn('ring must exist',
                      self.swift_handoffs.log.get_loglines('error'))
        self.swift_handoffs.log = FakeLogger()

        instance = {'devices': self.datadir, 'ring': self.ring}
        with mock.patch('os.path.isfile', return_value=False):
            self.swift_handoffs.check(instance)
        self.assertIn('ring must exist',
                      self.swift_handoffs.log.get_loglines('error'))
        self.swift_handoffs.log = FakeLogger()

        # granularity defaults to server. If specified it only allows either
        # server or drive. Anything else will be an error.
        instance = {'devices': self.datadir, 'ring': self.ring,
                    'granularity': 'something else'}
        self.swift_handoffs.check(instance)
        self.assertIn("granularity must be either 'server' or 'drive'",
                      self.swift_handoffs.log.get_loglines('error'))

    def setup_partitions(self, devices):
        for dev in devices:
            for part in devices[dev]:
                path = os.path.join(self.datadir, dev, 'objects', str(part))
                os.makedirs(path)

    @mock.patch('monasca_agent.collector.checks_d.swift_handoffs.Ring',
                MockRing)
    def test_all_paritions_in_correct_place(self):
        self.setup_partitions(self.expected_dev2part)
        instances = {'devices': self.datadir, 'ring': self.ring,
                     'granularity': 'device'}
        self.swift_handoffs.check(instances)

        self.assertTrue(self.swift_handoffs.gauge_called)
        for metric in ('swift.partitions.primary_count',
                       'swift.partitions.handoff_count'):
            # metric was called
            self.assertIn(metric, self.swift_handoffs.gauge_calls)

            # Each metric was called once per device, so 4 times.
            self.assertEqual(len(self.swift_handoffs.gauge_calls[metric]), 4)

        self.assertListEqual(
            self.swift_handoffs.gauge_calls['swift.partitions.primary_count'],
            [12, 12, 12, 12])
        self.assertListEqual(
            self.swift_handoffs.gauge_calls['swift.partitions.handoff_count'],
            [0, 0, 0, 0])

        # each device should be a device metric
        self.assertSetEqual(self.swift_handoffs.dimensions['device'],
                            {'sdb3', 'sdb2', 'sdb1', 'sdb4'})

    @mock.patch('monasca_agent.collector.checks_d.swift_handoffs.Ring',
                MockRing)
    def test_all_paritions_and_all_handoffs(self):

        for device in self.expected_dev2part:
            self.expected_dev2part[device].update(
                self.expected_handoffs[device])
        self.setup_partitions(self.expected_dev2part)
        instances = {'devices': self.datadir, 'ring': self.ring,
                     'granularity': 'device'}
        self.swift_handoffs.check(instances)

        self.assertTrue(self.swift_handoffs.gauge_called)
        for metric in ('swift.partitions.primary_count',
                       'swift.partitions.handoff_count'):
            # metric was called
            self.assertIn(metric, self.swift_handoffs.gauge_calls)

            # Each metric was called once per device, so 4 times.
            self.assertEqual(len(self.swift_handoffs.gauge_calls[metric]), 4)

        # all primaries were on each drive
        self.assertListEqual(
            self.swift_handoffs.gauge_calls['swift.partitions.primary_count'],
            [12, 12, 12, 12])
        # so were 4 handoffs
        self.assertListEqual(
            self.swift_handoffs.gauge_calls['swift.partitions.handoff_count'],
            [4, 4, 4, 4])

        # each device should be a device metric
        self.assertSetEqual(self.swift_handoffs.dimensions['device'],
                            {'sdb3', 'sdb2', 'sdb1', 'sdb4'})

    @mock.patch('monasca_agent.collector.checks_d.swift_handoffs.Ring',
                MockRing)
    def test_some_paritions_in_correct_no_handoffs(self):
        # Are parition will only be created on a drive if an object in that
        # partition has a been PUT into the cluster. So a partition missing
        # on a drive isn't bad (though in a realy cluster weird) but isn't
        # a failure.

        # let's remove a bunch of partitions from each cluster.
        for drive in self.expected_dev2part:
            self.expected_dev2part[drive].difference_update(
                list(self.expected_dev2part[drive])[:5])
        self.setup_partitions(self.expected_dev2part)
        instances = {'devices': self.datadir, 'ring': self.ring,
                     'granularity': 'device'}
        self.swift_handoffs.check(instances)

        self.assertTrue(self.swift_handoffs.gauge_called)
        for metric in ('swift.partitions.primary_count',
                       'swift.partitions.handoff_count'):
            # metric was called
            self.assertIn(metric, self.swift_handoffs.gauge_calls)

            # Each metric was called once per device, so 4 times.
            self.assertEqual(len(self.swift_handoffs.gauge_calls[metric]), 4)

        self.assertListEqual(
            self.swift_handoffs.gauge_calls['swift.partitions.primary_count'],
            [7, 7, 7, 7])
        self.assertListEqual(
            self.swift_handoffs.gauge_calls['swift.partitions.handoff_count'],
            [0, 0, 0, 0])

        # each device should be a device metric
        self.assertSetEqual(self.swift_handoffs.dimensions['device'],
                            {'sdb3', 'sdb2', 'sdb1', 'sdb4'})

    @mock.patch('monasca_agent.collector.checks_d.swift_handoffs.Ring',
                MockRing)
    def test_some_paritions_and_some_handoffs_less_devices(self):
        # Are parition will only be created on a drive if an object in that
        # partition has a been PUT into the cluster. So a partition missing
        # on a drive isn't bad (though in a realy cluster weird) but isn't
        # a failure.

        # let's remove a bunch of partitions from each cluster and 2 of the
        # devices
        for drive in 'sdb1', 'sdb4':
            self.expected_dev2part.pop(drive)

        for drive in self.expected_dev2part:
            self.expected_dev2part[drive].difference_update(
                list(self.expected_dev2part[drive])[:5])
            self.expected_dev2part[drive].update(
                list(self.expected_handoffs[drive])[:1])
        self.setup_partitions(self.expected_dev2part)
        instances = {'devices': self.datadir, 'ring': self.ring,
                     'granularity': 'device'}
        self.swift_handoffs.check(instances)

        self.assertTrue(self.swift_handoffs.gauge_called)
        for metric in ('swift.partitions.primary_count',
                       'swift.partitions.handoff_count'):
            # metric was called
            self.assertIn(metric, self.swift_handoffs.gauge_calls)

            # Each metric was called once per device, so 4 times.
            self.assertEqual(len(self.swift_handoffs.gauge_calls[metric]), 2)

        self.assertListEqual(
            self.swift_handoffs.gauge_calls['swift.partitions.primary_count'],
            [7, 7])
        self.assertListEqual(
            self.swift_handoffs.gauge_calls['swift.partitions.handoff_count'],
            [1, 1])

        # each device should be a device metric
        self.assertSetEqual(self.swift_handoffs.dimensions['device'],
                            {'sdb3', 'sdb2'})



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest
import time
import threading
import os

from nose.plugins.skip import SkipTest

from monasca_agent.common.aggregator import MetricsAggregator
from monasca_agent.statsd import Server
from monasca_agent.common.config import get_logging_config
from monasca_agent.collector.jmxfetch import JMXFetch


STATSD_PORT = 8121


class DummyReporter(threading.Thread):

    def __init__(self, metrics_aggregator):
        threading.Thread.__init__(self)
        self.finished = threading.Event()
        self.metrics_aggregator = metrics_aggregator
        self.interval = 10
        self.metrics = None
        self.finished = False
        self.start()

    def run(self):
        while not self.finished:
            time.sleep(self.interval)
            self.flush()

    def flush(self):
        metrics = self.metrics_aggregator.flush()
        if metrics:
            self.metrics = metrics


class JMXTestCase(unittest.TestCase):

    def setUp(self):
        aggregator = MetricsAggregator("test_host")
        self.server = Server(aggregator, "localhost", STATSD_PORT)
        self.reporter = DummyReporter(aggregator)

        self.t1 = threading.Thread(target=self.server.start)
        self.t1.start()

        confd_path = os.path.realpath(os.path.join(os.path.abspath(__file__), "..", "jmx_yamls"))
        JMXFetch.init(confd_path, {'dogstatsd_port': STATSD_PORT}, get_logging_config(), 15)

    def tearDown(self):
        self.server.stop()
        self.reporter.finished = True
        JMXFetch.stop()

    def testCustomJMXMetric(self):
        raise SkipTest('Requires JMX be setup')
        count = 0
        while self.reporter.metrics is None:
            time.sleep(1)
            count += 1
            if count > 20:
                raise Exception("No metrics were received in 20 seconds")

        metrics = self.reporter.metrics

        self.assertIsInstance(metrics, list)
        self.assertTrue(len(metrics) > 0)
        self.assertTrue(len([t for t in metrics if "cassandra.db." in t[
                        'metric'] and "instance:cassandra_instance" in t['dimensions']]) > 40, metrics)

if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2018 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os

import monasca_setup.agent_config
import monasca_setup.detection

LOG = logging.getLogger(__name__)

_IB_DEVICE_PATH = "/sys/class/infiniband/"


class IBNetworkDetect(monasca_setup.detection.Plugin):
    """Detects and configures Infiniband plugin."""
    def _detect(self):
        self.available = False
        if not self._detect_infiniband():
            LOG.info('Infiniband hardware was not detected: ib_network plugin'
                     'will not be loaded.')
            return
        self.available = True

    def build_config(self):
        config = monasca_setup.agent_config.Plugins()
        config['ib_network'] = {'init_config': None,
                                'instances': [{'name': 'ib_network_stats'}]}
        return config

    def _detect_infiniband(self):
        return os.path.isdir(_IB_DEVICE_PATH)



# Copyright 2018 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import monasca_setup.detection


class Cassandra(monasca_setup.detection.ServicePlugin):

    """Detect Cassandra daemons and setup configuration to monitor them.
        Cassandra directories can be checked by passing in a directory_names
        list. example: 'directory_names': [('/var/cassasndra/data',
        '/var/cassasndra/commitlog', '/var/log/cassandra')]
        Cassandra process user name can be overwritten by passing in
        process_username.
        See ServicePlugin for details.
    """

    def __init__(self, template_dir, overwrite=True, args=None):
        service_params = {
            'args': args,
            'template_dir': template_dir,
            'overwrite': overwrite,
            'service_name': 'cassandra',
            'component_name': 'cassandra',
            'process_username': 'cassandra',
            'service_api_url': '',
            'search_pattern': ''
        }

        super(Cassandra, self).__init__(service_params)



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ast

import monasca_setup.agent_config
import monasca_setup.detection


class HttpCheck(monasca_setup.detection.ArgsPlugin):
    """Setup an http_check according to the passed in args.
       Despite being a detection plugin this plugin does no detection and will be a noop without
       arguments.
       Expects space separated arguments, the required argument is url. Optional parameters include:
       disable_ssl_validation and match_pattern.

       You could provide Keystone configuration for all services when
       for example you want to monitor different OpenStack instance.
       Using monasca-setup, you can pass in space separated arguments with
       --detection_args, or the shortcut -a.

       monasca-setup -d HttpCheck --detection_args \
         "keystone_url=http://192.168.10.6/identity \
          keystone_project=proj \
          keystone_project_domain=Default \
          keystone_user=usr \
          keystone_user_domain=Default \
          keystone_password=pass \
          url=https://192.168.10.6"

       NOTE: keystone_project_domain and keystone_user_domain are required
       for Keystone V3. They are used to convey the project's domain name and
       user's domain name respectively.
    """

    def _detect(self):
        """Run detection, set self.available True if the service is detected.
        """
        self.available = self._check_required_args(['url'])

    def build_config(self):
        """Build the config as a Plugins object and return.
        """
        config = monasca_setup.agent_config.Plugins()
        # No support for setting headers at this time
        instance = self._build_instance(['url', 'timeout', 'username', 'password',
                                         'match_pattern', 'disable_ssl_validation',
                                         'name', 'use_keystone', 'collect_response_time'])

        init_config = None
        if 'keystone_url' in self.args:
            init_config = {'keystone_config': self._build_instance([
                'keystone_url',
                'keystone_project',
                'keystone_project_domain',
                'keystone_user',
                'keystone_user_domain',
                'keystone_password'],
                add_dimensions=False)}

        # Normalize any boolean parameters
        for param in ['use_keystone', 'collect_response_time']:
            if param in self.args:
                instance[param] = ast.literal_eval(self.args[param].capitalize())
        # Set some defaults
        if 'collect_response_time' not in instance:
            instance['collect_response_time'] = True
        if 'name' not in instance:
            instance['name'] = self.args['url']

        # Configure http check wide Keystone settings
        if init_config and ('keystone_config' in init_config):
            if 'keystone_project' in init_config['keystone_config']:
                init_config['keystone_config']['project_name'] = init_config['keystone_config']\
                    .pop('keystone_project')
            if 'keystone_project_domain' in init_config['keystone_config']:
                init_config['keystone_config']['project_domain_name'] = \
                    init_config['keystone_config'].pop('keystone_project_domain')
            if 'keystone_user' in init_config['keystone_config']:
                init_config['keystone_config']['username'] = init_config['keystone_config']\
                    .pop('keystone_user')
            if 'keystone_user_domain' in init_config['keystone_config']:
                init_config['keystone_config']['user_domain_name'] = \
                    init_config['keystone_config'].pop('keystone_user_domain')
            if 'keystone_password' in init_config['keystone_config']:
                init_config['keystone_config']['password'] = init_config['keystone_config']\
                    .pop('keystone_password')

        config['http_check'] = {'init_config': init_config, 'instances': [instance]}
        return config



# Copyright 2016 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from oslotest import base
import psutil

from monasca_setup.detection.plugins import zookeeper as zk

_DEFAULT_CFG_FILE = '/etc/zookeeper/conf/zoo.cfg'


def _get_cmd(config_file=_DEFAULT_CFG_FILE):
    return 'org.apache.zookeeper.server.quorum.QuorumPeerMain {0}'.format(config_file)


_ZOOKEEPER_CMD = _get_cmd()


class FakeProcess(object):
    cmdLine = None

    def as_dict(self, attrs=None):
        all_attrs = {'name': 'java',
                     'exe': FakeProcess.exe(),
                     'cmdline': FakeProcess.cmdline()}
        if attrs:
            for key in attrs:
                if key not in all_attrs:
                    all_attrs.pop(key, None)
        return all_attrs

    @staticmethod
    def exe():
        line = FakeProcess.cmdLine
        if not line:
            return None
        return line[0]

    @staticmethod
    def cmdline():
        return FakeProcess.cmdLine


class TestZookeeperDetection(base.BaseTestCase):

    FAKE_CONFIG_FILE_WITH_OPTIONS = ['tickTime=2000', 'initLimit=10',
                                     'syncLimit=5', 'dataDir=/var/lib/zookeeper',
                                     'clientPortAddress=192.168.10.6', 'clientPort=2182']

    FAKE_CONFIG_FILE_WITHOUT_OPTIONS = ['tickTime=2000', 'initLimit=10',
                                        'syncLimit=5', 'dataDir=/var/lib/zookeeper']

    BUILD_CONFIG = {'init_config': None,
                    'instances': [{'name': '192.168.10.6', 'host': '192.168.10.6',
                                   'port': 2181, 'timeout': 3}]}

    def setUp(self):
        super(TestZookeeperDetection, self).setUp()
        with mock.patch.object(zk.Zookeeper, '_detect') as mock_detect:
            self._zk = zk.Zookeeper('zookeeper')
            self.assertTrue(mock_detect.called)

    def test_should_not_configure_if_no_process(self):
        FakeProcess.cmdLine = []
        self._detect(proc=True)
        self.assertFalse(self._zk.available)

    def test_should_not_configure_has_process_no_config_located(self):
        FakeProcess.cmdLine = [_ZOOKEEPER_CMD]
        self._zk._get_config_file = mock.Mock(return_value=None)
        self._detect()
        self.assertFalse(self._zk.available)

    @mock.patch('monasca_setup.detection.plugins.zookeeper.os.path.isfile')
    def test_should_be_available_if_everything_matches(self, is_f):
        FakeProcess.cmdLine = [_ZOOKEEPER_CMD]
        is_f.return_value = True

        self._zk._get_config_file = mock.Mock(return_value=_DEFAULT_CFG_FILE)

        self._detect()
        self.assertTrue(self._zk.available)

    def test_should_detect_config_file_from_cmdline(self):
        FakeProcess.cmdLine = [_ZOOKEEPER_CMD]
        self.assertTrue(_DEFAULT_CFG_FILE, zk.Zookeeper._get_config_file(FakeProcess()))

    @mock.patch('monasca_setup.detection.plugins.zookeeper.os.path.isfile')
    def test_should_be_available_use_default_config_file(self, is_f):
        FakeProcess.cmdLine = [_get_cmd(config_file='')]
        is_f.return_value = True

        self._detect()
        self.assertTrue(self._zk.available)

    def test_should_return_default_config_file_if_in_cmdline_is_missing(self):
        FakeProcess.cmdLine = ['zookeeper']
        self.assertEqual('/etc/zookeeper/conf/zoo.cfg', zk.Zookeeper._get_config_file(FakeProcess()))

    @mock.patch('monasca_setup.detection.plugins.zookeeper.open')
    def test_should_return_options_from_config_file(self, open_file):
        open_file.return_value = self.FAKE_CONFIG_FILE_WITH_OPTIONS
        FakeProcess.cmdLine = []
        ip_address = '192.168.10.6'
        port = 2182
        self.assertEqual((ip_address, port), zk.Zookeeper._read_config_file(FakeProcess()))

    @mock.patch('monasca_setup.detection.plugins.zookeeper.open')
    def test_should_return_default_options_missing_in_config_file(self, open_file):
        open_file.return_value = self.FAKE_CONFIG_FILE_WITHOUT_OPTIONS
        FakeProcess.cmdLine = []
        ip_address = 'localhost'
        port = 2181
        self.assertEqual((ip_address, port), zk.Zookeeper._read_config_file(FakeProcess()))

    @mock.patch('monasca_setup.detection.plugins.zookeeper.open')
    def test_exception_while_parrsing_file(self, open_file):
        open_file.return_value = ['clientPort=aa']
        self.assertEqual(None, zk.Zookeeper._read_config_file(FakeProcess()))

    def test_should_build_config(self):
        config = ('192.168.10.6', 2181)
        FakeProcess.cmdLine = [_ZOOKEEPER_CMD]
        self._zk._get_config_file = mock.Mock(return_value=None)
        self._detect()
        self._zk._read_config_file = mock.Mock(return_value=config)
        self.assertEqual(self.BUILD_CONFIG, self._zk.build_config()['zk'])

    def _detect(self, proc=False):
        self._zk.available = False
        processes = [FakeProcess()] if not proc else []
        process_iter = mock.patch.object(psutil, 'process_iter',
                                         return_value=processes)
        with process_iter as mock_process_iter:
            self._zk._detect()
            self.assertTrue(mock_process_iter.called)



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


class Singleton(type):
    def __init__(cls, name, bases, dict):
        super(Singleton, cls).__init__(name, bases, dict)
        cls.instance = None

    def __call__(cls, *args, **kw):
        if cls.instance is None:
            cls.instance = super(Singleton, cls).__call__(*args, **kw)
        return cls.instance



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from oslotest import base
import psutil

from monasca_setup.detection.plugins import influxdb as idb

_DEFAULT_CFG_FILE = '/etc/influxdb/influxdb.conf'


def _get_cmd(config_file=_DEFAULT_CFG_FILE):
    """Builds mocked cmdline for process"""
    return ('/usr/bin/influxd -config %s' % config_file).split(' ')


_INFLUXDB_CMD = _get_cmd()


class FakeProcess(object):
    cmdLine = None

    def as_dict(self, attrs=None):
        all_attrs = {'name': 'influxd',
                     'exe': FakeProcess.exe(),
                     'cmdline': FakeProcess.cmdline()}
        if attrs:
            for key in attrs:
                if key not in all_attrs:
                    all_attrs.pop(key, None)
        return all_attrs

    @staticmethod
    def exe():
        line = FakeProcess.cmdLine
        if not line:
            return None
        return line[0]

    @staticmethod
    def cmdline():
        return FakeProcess.cmdLine


class TestInfluxDBDetection(base.BaseTestCase):
    ADDRESSES = {
        ':8086': ('127.0.0.1', 8086),
        '192.168.10.6:8888': ('192.168.10.6', 8888)
    }
    LOCATIONS = (
        '/tmp/influx.conf',
        _DEFAULT_CFG_FILE,
        '/etc/monasca/influx.conf'
    )

    def setUp(self):
        super(TestInfluxDBDetection, self).setUp()
        with mock.patch.object(idb.InfluxDB, '_detect') as mock_detect:
            self._ir = idb.InfluxDB('influxdb')
            self.assertTrue(mock_detect.called)

    def test_should_not_configure_if_no_process(self):
        FakeProcess.cmdLine = []  # no_process
        self._detect(no_proc=True)
        self.assertFalse(self._ir.available)

    def test_should_not_configure_has_process_no_config_located(self):
        FakeProcess.cmdLine = [_INFLUXDB_CMD]
        self._ir._get_config_file = mock.Mock(return_value=None)
        self._detect()
        self.assertFalse(self._ir.available)

    @mock.patch('monasca_setup.detection.plugins.influxdb.importutils')
    def test_should_not_configure_no_dependencies(self, iu):
        FakeProcess.cmdLine = [_INFLUXDB_CMD]
        self._ir._get_config_file = mock.Mock(return_value=True)
        iu.return_value = False
        self.assertFalse(self._ir.available)

    @mock.patch('monasca_setup.detection.plugins.influxdb.importutils')
    def test_should_be_available_if_everything_matches(self, iu):
        FakeProcess.cmdLine = [_INFLUXDB_CMD]

        self._ir._get_config_file = mock.Mock(return_value=_DEFAULT_CFG_FILE)
        self._ir._load_config = lc = mock.Mock()
        iu.try_import.return_value = True

        self._detect()

        self.assertTrue(self._ir.available)
        lc.assert_called_with(_DEFAULT_CFG_FILE)

    @mock.patch('monasca_setup.detection.plugins.influxdb.importutils')
    def test_dependencies_installed_true_has_toml(self, iu):
        iu.try_import = tr = mock.Mock(return_value=True)
        self.assertTrue(self._ir.dependencies_installed())
        tr.assert_called_with('toml', False)

    @mock.patch('monasca_setup.detection.plugins.influxdb.importutils')
    def test_dependencies_installed_false_no_toml(self, iu):
        iu.try_import = tr = mock.Mock(return_value=False)
        self.assertFalse(self._ir.dependencies_installed())
        tr.assert_called_with('toml', False)

    def test_should_explode_addresses(self):
        for raw_address, e_host_port in self.ADDRESSES.items():
            http_conf = {
                'bind-address': raw_address
            }
            a_host_port = idb.InfluxDB._explode_bind_address(http_conf)
            self.assertEqual(e_host_port, a_host_port)

    def test_should_return_none_cfg_file_if_cmd_switch_missing(self):
        FakeProcess.cmdLine = []
        self.assertIsNone(idb.InfluxDB._get_config_file(FakeProcess()))

    def test_should_return_cfg_file_path_if_cmd_switch_found(self):
        for loc in self.LOCATIONS:
            FakeProcess.cmdLine = _get_cmd(config_file=loc)
            self.assertEqual(loc,
                             idb.InfluxDB._get_config_file(FakeProcess()))

    @mock.patch('monasca_setup.detection.plugins.influxdb.'
                'utils.find_addrs_listening_on_port')
    def test_should_build_config_db_listens(self, falop):
        falop.return_value = True
        self._build_config(process_up=True)

    @mock.patch('monasca_setup.detection.plugins.influxdb.'
                'utils.find_addrs_listening_on_port')
    def test_should_build_config_db_died_during_conf(self, falop):
        falop.return_value = False
        self._build_config(process_up=False)

    @mock.patch('monasca_setup.detection.plugins.influxdb.'
                'utils.find_addrs_listening_on_port')
    def test_should_build_config_http_disabled(self, falop):
        falop.return_value = False
        self._build_config(http_enabled=False)

    @mock.patch('monasca_setup.detection.plugins.influxdb.'
                'utils.find_addrs_listening_on_port')
    def test_should_build_config_db_listens_with_influxdb_node_key(self, falop):
        falop.return_value = True
        self._ir.args = {'influxdb_node': 'test-key'}
        self._build_config(process_up=True)

    def _build_config(self, http_enabled=True, process_up=True):
        """Verify built configuration

        :param process_up: True/False, intermediate process availability
        :type process_up: bool
        :param http_enabled: Is http enabled for given influx
        :type http_enabled: bool

        """
        monitored_items = ['process']
        if process_up and http_enabled:
            monitored_items.append('http_check')

        for raw_address, host_port in self.ADDRESSES.items():
            self._ir._config = {
                'http': {
                    'enabled': http_enabled,
                    'bind-address': raw_address
                }
            }
            built_config = self._ir.build_config()

            self.assertCountEqual(monitored_items, built_config.keys())
            for key in built_config.keys():
                if key == 'process':
                    self._verify_process_conf(built_config[key])
                elif key == 'http_check':
                    self._verify_http_conf(built_config[key], host_port)
                else:
                    raise 'Untested monitored item %s' % key

    def _verify_http_conf(self, built_config, e_host_port):
        dimensions = {
            'component': 'influxdb',
            'service': 'influxdb'
        }
        if self._ir.args and self._ir.args.get(self._ir.INFLUXDB_NODE_ARG_NAME):
            dimensions.update(
                {self._ir.INFLUXDB_NODE_ARG_NAME: self._ir.args.get(self._ir.INFLUXDB_NODE_ARG_NAME)})

        expected_http = {
            'init_config': None,
            'instances': [
                {
                    'name': 'influxdb',
                    'url': 'http://%s:%d/ping' % e_host_port,
                    'dimensions': dimensions
                }
            ]
        }
        self.assertDictEqual(expected_http, built_config)

    def _verify_process_conf(self, actual_config):
        dimensions = {
            'component': 'influxdb',
            'service': 'influxdb'
        }
        if self._ir.args and self._ir.args.get(self._ir.INFLUXDB_NODE_ARG_NAME):
            dimensions.update(
                {self._ir.INFLUXDB_NODE_ARG_NAME: self._ir.args.get(self._ir.INFLUXDB_NODE_ARG_NAME)})

        expected_process = {
            'init_config': None,
            'instances': [
                {
                    'detailed': True,
                    'search_string': ['influxd'],
                    'exact_match': False,
                    'name': 'influxd',
                    'dimensions': dimensions
                }
            ]
        }
        self.assertDictEqual(expected_process, actual_config)

    def _detect(self, no_proc=False):
        self._ir.available = False
        processes = [FakeProcess()] if not no_proc else []
        process_iter = mock.patch.object(psutil, 'process_iter',
                                         return_value=processes)
        with process_iter as mock_process_iter:
            self._ir._detect()
            self.assertTrue(mock_process_iter.called)



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Windows Only.

Generic WMI check. This check allows you to specify particular metrics that you
want from WMI in your configuration. Check wmi.yaml.example in your conf.d
directory for more details on configuration.
"""
try:
    import wmi
except Exception:
    wmi = None

from monasca_agent.collector.checks import AgentCheck

UP_METRIC = 'Up'
SEARCH_WILDCARD = '*'


class WMICheck(AgentCheck):

    def __init__(self, name, init_config, agent_config):
        AgentCheck.__init__(self, name, init_config, agent_config)
        self.wmi_conns = {}

    def _get_wmi_conn(self, host, user, password):
        key = "%s:%s:%s" % (host, user, password)
        if key not in self.wmi_conns:
            self.wmi_conns[key] = wmi.WMI(computer=host, user=user,
                                          password=password)
        return self.wmi_conns[key]

    def check(self, instance):
        if wmi is None:
            raise Exception("Missing 'wmi' module")

        host = instance.get('host', None)
        user = instance.get('username', None)
        password = instance.get('password', None)
        w = self._get_wmi_conn(host, user, password)

        wmi_class = instance.get('class')
        metrics = instance.get('metrics')
        filters = instance.get('filters')
        tag_by = instance.get('tag_by')

        if not wmi_class:
            raise Exception('WMI instance is missing a value for `class` in wmi.yaml')

        # If there are filters, we need one query per filter.
        if filters:
            for f in filters:
                prop = list(f.keys())[0]
                search = list(f.values())[0]
                if SEARCH_WILDCARD in search:
                    search = search.replace(SEARCH_WILDCARD, '%')
                    wql = "SELECT * FROM %s WHERE %s LIKE '%s'" % (wmi_class, prop, search)
                    results = w.query(wql)
                else:
                    results = getattr(w, wmi_class)(**f)
                self._extract_metrics(results, metrics, tag_by, instance)
        else:
            results = getattr(w, wmi_class)()
            self._extract_metrics(results, metrics, tag_by, instance)

    def _extract_metrics(self, results, metrics, tag_by, instance):
        if len(results) > 1 and tag_by is None:
            raise Exception(
                'WMI query returned multiple rows but no `tag_by` value was given. metrics=%s' %
                metrics)

        for wmi_property, name, mtype in metrics:
            for res in results:
                if wmi_property == UP_METRIC:
                    # Special-case metric will just submit 1 for every value
                    # returned in the result.
                    val = 1
                else:
                    val = float(getattr(res, wmi_property))

                # Grab the tag from the result if there's a `tag_by` value (e.g.: "name:jenkins")
                if tag_by:
                    dimensions = {'{0}'.format(tag_by.lower()): getattr(res, tag_by)}
                else:
                    dimensions = None

                try:
                    func = getattr(self, mtype)
                except AttributeError:
                    raise Exception('Invalid metric type: {0}'.format(mtype))

                # submit the metric
                func(name, val, dimensions=self._set_dimensions(dimensions, instance))



# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

import monasca_setup.agent_config
import monasca_setup.detection

log = logging.getLogger(__name__)


class HostAlive(monasca_setup.detection.ArgsPlugin):
    """Setup an host_alive check according to the passed in args.
       Despite being a detection plugin, this plugin does no detection and
       will be a NOOP without arguments.  Expects two space-separated
       arguments, 'hostname' and 'type,' where the former is a comma-separated
       list of hosts, and the latter can be either 'ssh' or 'ping'.
       Examples:

       monasca-setup -d hostalive -a "hostname=remotebox type=ping"

       monasca-setup -d hostalive -a "hostname=rb,rb2 target_hostname=,rb2-nic2 type=ssh"
    """

    DEFAULT_PING_TIMEOUT = 1
    DEFAULT_SSH_TIMEOUT = 2
    DEFAULT_SSH_PORT = 22

    def _detect(self):
        """Run detection, set self.available True if the service is detected.
        """
        self.available = self._check_required_args(['hostname', 'type'])
        # Ideally, the arg would be called 'hostnames,' but leaving it
        # 'hostname' avoids breaking backward compatibility.

    def build_config(self):
        """Build the config as a Plugins object and return.
        """
        config = monasca_setup.agent_config.Plugins()
        log.info("\tEnabling {type} host check for {hostname}".format(**self.args))
        instances = []
        for hostname in self.args['hostname'].split(','):
            # Since the naming in the args and in the config don't match,
            #  build_instance is only good for dimensions
            instance = self._build_instance([])
            instance.update({'name': "{0} {1}".format(hostname,
                                                      self.args['type']),
                             'host_name': hostname,
                             'alive_test': self.args['type']})
            instances.append(instance)
        if 'target_hostname' in self.args:
            index = 0
            network_names_to_check = self.args['target_hostname'].split(',')
            for target_hostname in network_names_to_check:
                if target_hostname:
                    if index >= len(instances):
                        raise Exception('Too many target_hostname values')
                    instance = instances[index]
                    instance.update({'target_hostname': target_hostname})
                index += 1

        config['host_alive'] = {
            'init_config': {'ping_timeout': self.DEFAULT_PING_TIMEOUT,
                            'ssh_timeout': self.DEFAULT_SSH_TIMEOUT,
                            'ssh_port': self.DEFAULT_SSH_PORT},
            'instances': instances}

        return config



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import monasca_setup.agent_config
import monasca_setup.detection

log = logging.getLogger(__name__)


class PowerDNS(monasca_setup.detection.Plugin):

    """Detect PowerDNS daemon and setup configuration to monitor it.

    """

    def _detect(self):
        """Run detection, set self.available True if the service is detected.

        """
        if monasca_setup.detection.find_process_cmdline('pdns_server') is not None:
            self.available = True

    def build_config(self):
        """Build the config and return.

        """
        log.info("\tWatching the pdns_server process.")
        return monasca_setup.detection.watch_process(
            ['pdns_server'], 'powerdns', exact_match=False)

    def dependencies_installed(self):
        return True
